"""netconf.py module is a wrapper around the ncclient package."""

import re
import atexit
import logging
import subprocess
import datetime
import lxml.etree as et
from time import sleep
from threading import Thread, Event
from ncclient import manager
from ncclient import operations
from ncclient import transport
from ncclient.operations.retrieve import GetReply
from ncclient.devices.default import DefaultDeviceHandler
from ncclient.operations.errors import TimeoutExpiredError

try:
    from pyats.connections import BaseConnection
    from pyats.utils.secret_strings import to_plaintext
    from pyats.log.utils import banner
except ImportError:
    class BaseConnection:
        pass

# create a logger for this module
logger = logging.getLogger(__name__)

nccl = logging.getLogger("ncclient")
# The 'Sending' messages are logged at level INFO.
# The 'Received' messages are logged at level DEBUG.


class NetconfSessionLogHandler(logging.Handler):
    """Logging handler that pretty prints ncclient XML."""

    parser = et.XMLParser(recover=True)

    def emit(self, record):
        if hasattr(record, 'session'):
            try:
                # If the message contains XML, pretty-print it
                record.args = list(record.args)

                for i in range(len(record.args)):
                    try:
                        arg = None
                        if isinstance(record.args[i], str):
                            arg = record.args[i].encode("utf-8")
                        elif isinstance(record.args[i], bytes):
                            arg = record.args[i]
                        if not arg:
                            continue
                        start = arg.find(b"<")
                        end = arg.rfind(b"]]>]]>")   # NETCONF 1.0 terminator
                        if end == -1:
                            end = arg.rfind(b">")
                            if end != -1:
                                # Include the '>' character in our range
                                end += 1
                        if start != -1 and end != -1:
                            elem = et.fromstring(arg[start:end], self.parser)
                            if elem is None:
                                continue

                            text = et.tostring(elem, pretty_print=True,
                                               encoding="utf-8")
                            record.args[i] = (arg[:start] +
                                              text +
                                              arg[end:]).decode()
                    except Exception:
                        # Pretty print issue so leave record unchanged
                        continue

                record.args = tuple(record.args)
            except Exception:
                # Unable to handle record so leave it unchanged
                pass


nccl.addHandler(NetconfSessionLogHandler())


class Netconf(manager.Manager, BaseConnection):
    '''Netconf

    Implementation of NetConf connection to devices (NX-OS, IOS-XR or IOS-XE),
    based on pyATS BaseConnection and ncclient.

    YAML Example::

        devices:
            asr22:
                type: 'ASR'
                tacacs:
                    login_prompt: "login:"
                    password_prompt: "Password:"
                    username: "admin"
                passwords:
                    tacacs: admin
                    enable: admin
                    line: admin
                connections:
                    a:
                        protocol: telnet
                        ip: "1.2.3.4"
                        port: 2004
                    vty:
                        protocol : telnet
                        ip : "2.3.4.5"
                    netconf:
                        class: yang.connector.Netconf
                        ip : "2.3.4.5"
                        port: 830
                        username: admin
                        password: admin

    Code Example::

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/asr22.yaml')
        >>> device = testbed.devices['asr22']
        >>> device.connect(alias='nc', via='netconf')
        >>> device.nc.connected
        True
        >>> netconf_request = """
        ...     <rpc message-id="101"
        ...      xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
        ...     <get>
        ...     <filter>
        ...     <native xmlns="http://cisco.com/ns/yang/ned/ios">
        ...     <version>
        ...     </version>
        ...     </native>
        ...     </filter>
        ...     </get>
        ...     </rpc>
        ...     """
        >>> reply = device.nc.request(netconf_request)
        >>> print(reply)
        <?xml version="1.0" encoding="UTF-8"?>
        <rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"
        message-id="101"><data>
        <native xmlns="http://cisco.com/ns/yang/ned/ios">
        <version>16.3</version></native></data></rpc-reply>
        >>> device.nc.disconnect()
        >>> device.nc.connected
        False
        >>>

    Attributes
    ----------
    timeout : `int`
        Timeout value in seconds which is used by paramiko channel. By
        default this value is 30 seconds.

    client_capabilities : `object`
        Object ncclient.capabilities.Capabilities representing the client's
        capabilities.

    server_capabilities : `object`
        Object ncclient.capabilities.Capabilities representing the server's
        capabilities, and it has a list of data models the server supports.

    async_mode : `boolean`
        Specify whether operations are executed asynchronously (True) or
        synchronously (False). The default value is False.
    '''

    def __init__(self, *args, **kwargs):
        '''
        __init__ instantiates a single connection instance.
        '''
        # set defaults
        kwargs.setdefault('timeout', 30)

        # instanciate BaseConnection
        # (could use super...)
        BaseConnection.__init__(self, *args, **kwargs)

        # shortwire Ncclient device handling portion
        # and create just the DeviceHandler
        device_handler = DefaultDeviceHandler()

        # create the session instance
        session = transport.SSHSession(device_handler)

        # load known_hosts file (if available)
        if kwargs.get('hostkey_verify'):
            session.load_known_hosts()

        # instanciate ncclient Manager
        # (can't use super due to mro change)
        manager.Manager.__init__(
            self, session=session, device_handler=device_handler,
            timeout=self.timeout)

        self.active_notifications = {}

    @property
    def session(self):
        '''session

        High-level api: return the SSH session object.

        Returns
        -------

        object
            The SSH session that was created by ncclient.transport.SSHSession.
        '''

        return self._session

    def connect(self):
        '''connect

        High-level api: opens the NetConf connection and exchanges
        capabilities. Since topology YAML file is parsed by BaseConnection,
        the following parameters can be specified in your YAML file.

        Parameters
        ----------

        host : `string`
            Hostname or IP address to connect to.
        port : `int`, optional
            By default port is 830, but some devices use the default SSH port
            of 22 so this may need to be specified.
        timeout : `int`, optional
            An optional keyed argument to set timeout value in seconds. By
            default this value is 30 seconds.
        username : `string`
            The username to use for SSH authentication.
        password : `string`
            The password used if using password authentication, or the
            passphrase to use for unlocking keys that require it.
        key_filename : `string`
            a filename where a the private key to be used can be found.
        allow_agent : `boolean`
            Enables querying SSH agent (if found) for keys. The default value
            is True.
        hostkey_verify : `boolean`
            Enables hostkey verification from ~/.ssh/known_hosts. The default
            value is False.
        look_for_keys : `boolean`
            Enables looking in the usual locations for ssh keys
            (e.g. ~/.ssh/id_*). The default value is True.
        ssh_config : `string`
            Enables parsing of an OpenSSH configuration file, if set to its
            path, e.g. ~/.ssh/config or to True. If the value is True,
            ncclient uses ~/.ssh/config. The default value is None.

        Raises
        ------

        Exception
            If the YAML file does not have correct connections section, or
            establishing transport to ip:port is failed, ssh authentication is
            failed, or other transport failures.

        Note
        ----

        There is no return from this method. If something goes wrong, an
        exception will be raised.


        YAML Example::

            devices:
                asr22:
                    type: 'ASR'
                    tacacs:
                        login_prompt: "login:"
                        password_prompt: "Password:"
                        username: "admin"
                    passwords:
                        tacacs: admin
                        enable: admin
                        line: admin
                    connections:
                        a:
                            protocol: telnet
                            ip: "1.2.3.4"
                            port: 2004
                        vty:
                            protocol : telnet
                            ip : "2.3.4.5"
                        netconf:
                            class: yang.connector.Netconf
                            ip : "2.3.4.5"
                            port: 830
                            username: admin
                            password: admin

        Code Example::

            >>> from pyats.topology import loader
            >>> testbed = loader.load('/users/xxx/xxx/asr22.yaml')
            >>> device = testbed.devices['asr22']
            >>> device.connect(alias='nc', via='netconf')
            >>>

        Expected Results::

            >>> device.nc.connected
            True
            >>> for iter in device.nc.server_capabilities:
            ...     print(iter)
            ...
            urn:ietf:params:xml:ns:yang:smiv2:RFC-1215?module=RFC-1215
            urn:ietf:params:xml:ns:yang:smiv2:SNMPv2-TC?module=SNMPv2-TC
            ...
            >>>
        '''

        if self.connected:
            return

        logger.debug(self.session)
        if not self.session.is_alive():
            self._session = transport.SSHSession(self._device_handler)

        # default values
        defaults = {
            'host': None,
            'port': 830,
            'timeout': 30,
            'username': None,
            'password': None,
            'key_filename': None,
            'allow_agent': False,
            'hostkey_verify': False,
            'look_for_keys': False,
            'ssh_config': None,
            }
        defaults.update(self.connection_info)

        # remove items
        disregards = ['class', 'model', 'protocol',
                      'async_mode', 'raise_mode', 'credentials']
        defaults = {k: v for k, v in defaults.items() if k not in disregards}

        # rename ip -> host, cast to str type
        if 'ip' in defaults:
            defaults['host'] = str(defaults.pop('ip'))

        # rename user -> username
        if 'user' in defaults:
            defaults['username'] = str(defaults.pop('user'))

        # check credentials
        if self.connection_info.get('credentials'):
            try:
                defaults['username'] = str(
                    self.connection_info['credentials']['netconf']['username'])
            except Exception:
                pass
            try:
                defaults['password'] = to_plaintext(
                    self.connection_info['credentials']['netconf']['password'])
            except Exception:
                pass

        # support sshtunnel
        if 'sshtunnel' in defaults:
            from unicon.sshutils import sshtunnel
            try:
                tunnel_port = sshtunnel.auto_tunnel_add(self.device, self.via)
                if tunnel_port:
                    defaults['host'] = self.device.connections[self.via] \
                                           .sshtunnel.tunnel_ip
                    defaults['port'] = tunnel_port
            except AttributeError as err:
                raise AttributeError("Cannot add ssh tunnel. \
                Connection %s may not have ip/host or port.\n%s"
                                     % (self.via, err))
            del defaults['sshtunnel']

        defaults = {k: getattr(self, k, v) for k, v in defaults.items()}

        try:
            self.session.connect(**defaults)
            logger.info('NETCONF CONNECTED')
        except Exception:
            if self.session.transport:
                self.session.close()
            raise

        @atexit.register
        def cleanup():
            if self.session.transport:
                self.session.close()

    def disconnect(self):
        '''disconnect

        High-level api: closes the NetConf connection.
        '''

        self.session.close()

    def subscribe(self, request):
        """ Creates a notification listener and mark it as active """
        notifier = Notification(self, request=request)
        notifier.start()
        if request['format']['request_mode'] == 'ON_CHANGE':
            # Get ready for trigger event
            notifier.event_triggered = True
        self.active_notifications[self] = notifier

    def notify_wait(self, steps):
        """ Activate notification listener and check results """
        notifier = self.active_notifications.get(self)
        if notifier:
            if steps.result.code != 1:
                notifier.stop()
                del self.active_notifications[self]
                return
            notifier.event_triggered = True
            logger.info(banner('NOTIFICATION EVENT TRIGGERED'))
            wait_for_sample = notifier.sample_interval - 1
            cntr = 1.0
            while cntr < float(notifier.stream_max):
                logger.info('Listening for notifications from subscribe stream, {} seconds elapsed'.format(
                    cntr)
                )
                cntr += 1
                if notifier.result is not None and wait_for_sample <= 0:
                    notifier.stop()
                    if notifier.result is True:
                        steps.passed(
                            '\n' + banner('NOTIFICATION RESPONSE PASSED')
                        )
                    else:
                        steps.failed(
                            '\n' + banner('NOTIFICATION RESPONSE FAILED')
                        )
                    break
                sleep(1)
                wait_for_sample -= 1
            else:
                notifier.stop()
                steps.failed(
                    '\n' + banner('STREAM TIMED OUT WITHOUT RESPONSE')
                )

            if self in self.active_notifications:
                del self.active_notifications[self]

    def configure(self, msg):
        '''configure

        High-level api: configure is a common method of console, vty and ssh
        sessions, however it is not supported by this Netconf class. This is
        just a placeholder in case someone mistakenly calls config method in a
        netconf session. An Exception is thrown out with explanation.

        Parameters
        ----------

        msg : `str`
            Any config CLI need to be sent out.

        Raises
        ------

        Exception
            configure is not a supported method of this Netconf class.
        '''

        raise Exception('configure is not a supported method of this Netconf '
                        'class, since a more suitable method, edit_config, is '
                        'recommended. There are nine netconf operations '
                        'defined by RFC 6241, and edit-config is one of them. '
                        'Also users can build any netconf requst, including '
                        'invalid netconf requst as negative test cases, in '
                        'XML format and send it by method request.')

    def execute(self, operation, *args, **kwargs):
        '''execute

        High-level api: The fact that most connection classes implement
        execute method lead us to add this method here as well.
        Supported operations are get, get_config, get_schema, dispatch,
        edit_config, copy_config, validate, commit, discard_changes,
        delete_config, lock, unlock, close_session, kill_session,
        poweroff_machine and reboot_machine. Refer to ncclient document for
        more details.
        '''

        # allow for operation string type
        if type(operation) is str:
            try:
                cls = manager.OPERATIONS[operation]
            except KeyError:
                raise ValueError('No such operation "%s".\n'
                                 'Supported operations are: %s' %
                                 (operation, list(manager.OPERATIONS.keys())))
        else:
            cls = operation

        time1 = datetime.datetime.now()
        reply = super().execute(cls, *args, **kwargs)
        time2 = datetime.datetime.now()
        reply.elapsed = time2 - time1
        return reply

    def request(self, msg, timeout=30, return_obj=False):
        '''request

        High-level api: sends message through NetConf session and returns with
        a reply. Exception is thrown out either the reply is in wrong
        format or timout. Users can modify timeout value (in seconds) by
        passing parameter timeout. Users may want to set a larger timeout when
        making a large query.

        Parameters
        ----------

        msg : `str`
            Any message need to be sent out in XML format. The message can be
            in wrong format if it is a negative test case. Because ncclient
            tracks same message-id in both rpc and rpc-reply, missing
            message-id in your rpc may cause exception when receiving
            rpc-reply. Most other wrong format rpc's can be sent without
            exception.
        timeout : `int`, optional
            An optional keyed argument to set timeout value in seconds. Its
            default value is 30 seconds.
        return_obj : `boolean`, optional
            Normally a string is returned as a reply. In other cases, we may
            want to return a RPCReply object, so we can access some attributes,
            e.g., reply.ok or reply.elapsed.

        Returns
        -------

        str or RPCReply
            The reply from the device in string. If something goes wrong, an
            exception will be raised. If return_obj=True, the reply is a
            RPCReply object.


        Raises
        ------

        Exception
            If NetConf is not connected, or there is a timeout when receiving
            reply.


        Code Example::

            >>> from pyats.topology import loader
            >>> testbed = loader.load('/users/xxx/xxx/asr_20_22.yaml')
            >>> device = testbed.devices['asr22']
            >>> device.connect(alias='nc', via='netconf')
            >>> netconf_request = """
            ...     <rpc message-id="101"
            ...      xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
            ...     <get>
            ...     <filter>
            ...     <native xmlns="http://cisco.com/ns/yang/Cisco-IOS-XE-native">
            ...     <version>
            ...     </version>
            ...     </native>
            ...     </filter>
            ...     </get>
            ...     </rpc>
            ...     """
            >>> reply = device.nc.request(netconf_request)
            >>>

        Expected Results::

            >>> print(reply)
            <?xml version="1.0" encoding="UTF-8"?>
            <rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"
            message-id="101"><data>
            <native xmlns="http://cisco.com/ns/yang/ned/ios">
            <version>16.3</version></native></data></rpc-reply>
            >>>
        '''

        rpc = RawRPC(session=self.session,
                     device_handler=self._device_handler,
                     timeout=timeout,
                     raise_mode=operations.rpc.RaiseMode.NONE)

        # identify message-id
        m = re.search(r'message-id="([A-Za-z0-9_\-:# ]*)"', msg)
        if m:
            rpc._id = m.group(1)
            rpc._listener.register(rpc._id, rpc)
            logger.debug(
                'Found message-id="%s" in your rpc, which is good.', rpc._id)
        else:
            logger.warning('Cannot find message-id in your rpc. You may '
                           'expect an exception when receiving rpc-reply '
                           'due to missing message-id.')

        if return_obj:
            return rpc._request(msg)
        else:
            return rpc._request(msg).xml

    def __getattr__(self, method):
        # avoid the __getattr__ from Manager class
        if hasattr(manager, 'VENDOR_OPERATIONS') and method \
                in manager.VENDOR_OPERATIONS or method in manager.OPERATIONS:
            return super().__getattr__(method)
        else:
            raise AttributeError("'%s' object has no attribute '%s'"
                                 % (self.__class__.__name__, method))


class NetconfEnxr():
    """Subclass using POSIX pipes to Communicate NETCONF messaging."""

    chunk = re.compile('(\n#+\\d+\n)')
    rpc_pipe_err = """
        <rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
        <rpc-error>
            <error-type>transport</error-type>
            <error-tag>resource-denied</error-tag>
            <error-severity>error</error-severity>
            <error-message>No pipe data returned</error-message>
        </rpc-error>
        </rpc-reply>"""

    def __init__(self, *args, **kwargs):
        self.manager = None
        self.proc = None
        self.buf = None
        self.server_capabilities = None

    def get_rpc(self, elements):
        """Return string representation of lxml element with rpc."""
        rpc_element = et.Element(
            'rpc',
            attrib={'message-id': '101'},
            nsmap={None: "urn:ietf:params:xml:ns:netconf:base:1.0"}
        )
        rpc_element.append(elements)
        return et.tostring(rpc_element,
                           pretty_print=True).decode()

    def recv_data(self):
        """Retrieve data from process pipe."""
        if not self.proc:
            logger.info('Not connected.')
        else:
            buf = ''
            while True:
                # TODO: Could be better...1 byte at a time...
                # but, too much buffer and it deadlocks!!
                data = self.proc.stdout.read(1)

                if not data:
                    return GetReply(self.rpc_pipe_err)

                buf += data

                if buf.endswith('\n##'):
                    buf = buf[:-3]
                    break

            logger.info(buf)
            buf = buf[buf.find('<'):]
            reply = re.sub(self.chunk, '', buf)
            return GetReply(reply)

    def request(self, rpc):
        return self.send_cmd(rpc)

    def configure(self, msg):
        '''configure

        High-level api: configure is a common method of console, vty and ssh
        sessions, however it is not supported by this NetconfEnxr class. This is
        just a placeholder in case someone mistakenly calls config method in a
        netconf session. An Exception is thrown out with explanation.

        Parameters
        ----------

        msg : `str`
            Any config CLI need to be sent out.

        Raises
        ------

        Exception
            configure is not a supported method of this Netconf class.
        '''

        raise Exception('configure is not a supported method of this NetconfEnxr '
                        'class, since a more suitable method, edit_config, is '
                        'recommended. There are nine netconf operations '
                        'defined by RFC 6241, and edit-config is one of them. '
                        'Also users can build any netconf requst, including '
                        'invalid netconf requst as negative test cases, in '
                        'XML format and send it by method request.')

    def send_cmd(self, rpc):
        """Send a message to process pipe."""
        if not self.proc:
            logger.info('Not connected.')
        else:
            if et.iselement(rpc):
                if not rpc.tag.endswith('rpc'):
                    rpc = self.get_rpc(rpc)
                else:
                    rpc = et.tostring(rpc, pretty_print=True).decode()
            rpc_str = '\n#' + str(len(rpc)) + '\n' + rpc + '\n##\n'
            logger.info(rpc_str)
            self.proc.stdin.write(rpc_str)
            self.proc.stdin.flush()

            return self.recv_data()

    def edit_config(self, target=None, config=None, **kwargs):
        """Send edit-config."""
        target = target
        config = config
        target_element = et.Element('target')
        et.SubElement(target_element, target)
        edit_config_element = et.Element('edit-config')
        edit_config_element.append(target_element)
        edit_config_element.append(config)
        return self.send_cmd(self.get_rpc(edit_config_element))

    def get_config(self, source=None, filter=None, **kwargs):
        """Send get-config."""
        source = source
        filter = filter
        source_element = et.Element('source')
        et.SubElement(source_element, source)
        get_config_element = et.Element('get-config')
        get_config_element.append(source_element)
        get_config_element.append(filter)
        return self.send_cmd(self.get_rpc(get_config_element))

    def get(self, filter=None, **kwargs):
        filter_arg = filter
        get_element = et.Element('get')
        if isinstance(filter_arg, tuple):
            type, filter_content = filter_arg
            if type == "xpath":
                get_element.attrib["select"] = filter_content
            elif type == "subtree":
                filter_element = et.Element('filter')
                filter_element.append(filter_content)
                get_element.append(filter_element)
        else:
            get_element.append(filter_arg)
        return self.send_cmd(self.get_rpc(get_element))

    def commit(self, **kwargs):
        commit_element = et.Element('commit')
        return self.send_cmd(self.get_rpc(commit_element))

    def discard_changes(self, **kwargs):
        discard_element = et.Element('discard-changes')
        return self.send_cmd(self.get_rpc(discard_element))

    def lock(self, target=None, **kwargs):
        target = target
        store_element = et.Element(target)
        target_element = et.Element('target')
        target_element.append(store_element)
        lock_element = et.Element('lock')
        lock_element.append(target_element)
        return self.send_cmd(self.get_rpc(lock_element))

    def unlock(self, target=None, **kwargs):
        target = target
        store_element = et.Element(target)
        target_element = et.Element('target')
        target_element.append(store_element)
        unlock_element = et.Element('unlock')
        unlock_element.append(target_element)
        return self.send_cmd(self.get_rpc(unlock_element))

    def dispatch(self, rpc_command=None, **kwargs):
        rpc = rpc_command
        return self.send_cmd(rpc)

    @property
    def connected(self):
        """Check for active connection."""

        return self.server_capabilities is not None and self.proc.poll() \
            is None

    def connect(self, timeout=None):
        """Connect to ENXR pipe."""
        if self.connected:
            msg = 'Already connected'

        CMD = ['netconf_sshd_proxy', '-i', '0', '-o', '1', '-u', 'lab']
        BUFSIZE = 8192

        p = subprocess.Popen(CMD, bufsize=BUFSIZE,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             universal_newlines=True)

        buf = ''
        try:
            while True:
                data = p.stdout.read(1)
                if not data:
                    logger.info('No data received for hello')
                    p.terminate()
                    return

                buf += data
                if buf.endswith(']]>]]>'):
                    buf = buf[buf.find('<'):-6]
                    logger.info('Hello received')
                    break

            p.stdin.write(
                '<?xml version="1.0" encoding="UTF-8"?><hello '
                'xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"><capabilities>'
                '<capability>urn:ietf:params:netconf:base:1.1</capability>'
                '</capabilities></hello>]]>]]>'
            )
            p.stdin.flush()
            self.proc = p
            self.buf = ''
            elements = et.fromstring(buf)
            self.server_capabilities = [e.text for e in elements.iter()
                                        if hasattr(e, 'text')]
            # TODO: Notification stream interferes with get-schema
            msg = "NETCONF CONNECTED PIPE"
        except:
            msg = 'Not connected, Something went wrong'
        return msg

    def disconnect(self):
        """Disconnect from ENXR pipe."""
        if self.connected:
            self.proc.terminate()
            logger.info("NETCONF DISCONNECT PIPE")


class RawRPC(operations.rpc.RPC):
    '''RawRPC

    A modified ncclient.operations.rpc.RPC class. This is for internal use
    only.
    '''

    def _request(self, msg):
        '''_request

        Override method _request in class ncclient.operations.RPC, so it can
        handle raw rpc requests in string format without validating your rpc
        request syntax. When your rpc-reply is received, in most cases, it
        simply returns rpc-reply again in string format, except one scenario:
        If message-id is missing or message-id received does not match that in
        rpc request, ncclient will raise an OperationError.
        '''

        logger.debug('Requesting %r' % self.__class__.__name__)
        logger.info('Sending rpc...')
        logger.info(msg)
        time1 = datetime.datetime.now()
        self._session.send(msg)
        if not self._async:
            logger.debug('Sync request, will wait for timeout=%r' %
                         self._timeout)
            self._event.wait(self._timeout)
            if self._event.isSet():
                time2 = datetime.datetime.now()
                self._reply.elapsed = time2 - time1
                logger.info('Receiving rpc-reply after {:.3f} sec...'.
                            format(self._reply.elapsed.total_seconds()))
                logger.info(self._reply)
                return self._reply
            else:
                logger.info('Timeout. No rpc-reply received.')
                raise TimeoutExpiredError('ncclient timed out while waiting '
                                          'for an rpc-reply.')


class Notification(Thread):
    """ Listens for notifications, decodes, and verifies if any exists """
    def __init__(self, device, **request):
        Thread.__init__(self)
        self.device = device
        self.log = logging.getLogger(__name__)
        self.log.setLevel(logging.DEBUG)
        self._stop_event = Event()
        self.request = request
        self._event_triggered = False
        self._stopped = False

    @property
    def event_triggered(self):
        return self._event_triggered

    @event_triggered.setter
    def event_triggered(self, event_triggered):
        self._event_triggered = event_triggered

    @property
    def request(self):
        return self._request

    @request.setter
    def request(self, request={}):
        """ Sets the request property and propagates request's properties to the class """
        request_data = request['request']
        self.returns = request_data.get('returns')
        self.response_verify = request_data.get('verifier')
        self.decode_response = request_data.get('decode')
        self.namespace = request_data.get('namespace')
        self.sub_mode = request_data['format'].get('sub_mode', 'SAMPLE')
        self.encoding = request_data['format'].get('encoding', 'PROTO')
        self.sample_interval = request_data['format'].get('sample_interval', 10)
        if self.sub_mode == 'ON_CHANGE':
            self.sample_interval = 0
        self.stream_max = request_data['format'].get('stream_max', 0)
        self.time_delta = 0
        self.result = None
        self._event_triggered = False
        self._request = request_data

    def run(self):
        """ Start taking notifications until subscribe stream times out."""
        t1 = datetime.datetime.now()
        t2 = datetime.datetime.now()
        td = t2 - t1
        # Wait until after first sample period if sampling
        wait_for_sample = self.sample_interval - 1

        try:
            while self.time_delta < self.stream_max:
                t2 = datetime.datetime.now()
                td = t2 - t1

                if self.stopped():
                    self.time_delta = self.stream_max
                    self.log.info("Terminating notification thread")
                    break
                if self.stream_max:
                    t2 = datetime.datetime.now()
                    td = t2 - t1
                    self.time_delta = td.seconds
                    if td.seconds > self.stream_max:
                        self.stop()
                        break

                notif = self.device.take_notification(timeout=1)

                if notif and wait_for_sample <= 0:
                    resp_elements = self.decode_response(
                        notif.notification_xml
                    )
                    if resp_elements and self.returns:
                        self.result = self.response_verify(
                            resp_elements, self.returns
                        )
                        self.stop()
                        break
                wait_for_sample -= 1
        except Exception as exc:
            self.result = str(exc)
            self.log.error(str(exc))

    def stop(self):
        self.log.info("Stopping notification stream")
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()



import os
import logging
from collections import OrderedDict, Iterable
import base64
import json
from threading import Thread, Event
from time import sleep
from datetime import datetime
from six import string_types
from xml.etree.ElementPath import xpath_tokenizer_re

from google.protobuf import json_format

from cisco_gnmi import ClientBuilder
from cisco_gnmi import client

from . import xpath_util

try:
    from pyats.log.utils import banner
    from pyats.connections import BaseConnection
    from pyats.utils.secret_strings import to_plaintext
except ImportError:
    # Standalone without pyats install
    class BaseConnection:
        class dev:
            def __init__(self, dev_os):
                self.os = dev_os

        def __init__(self, device_os, **kwargs):
            self.connection_info = {'protocol': 'gnmi'}
            self._device = self.dev(device_os)
            self.connection_info.update(kwargs)

    def banner(string):
        return string

    def to_plaintext(string):
        return string

# create a logger for this module
log = logging.getLogger(__name__)
# Set debug level for cisco_gnmi.client
client.logger.setLevel(logging.DEBUG)


class gNMIException(IOError):
    pass


class GnmiNotification(Thread):
    """Thread listening for event notifications from the device."""

    def __init__(self, device, response, **request):
        Thread.__init__(self)
        self.device = device
        self._stop_event = Event()
        self.log = logging.getLogger(__name__)
        self.log.setLevel(logging.DEBUG)
        self.request = request
        self.responses = response

    @property
    def request(self):
        return self._request

    @request.setter
    def request(self, request={}):
        self.returns = request.get('returns')
        self.response_verify = request.get('verifier')
        self.decode_response = request.get('decode')
        self.namespace = request.get('namespace')
        self.sub_mode = request['format'].get('sub_mode', 'SAMPLE')
        self.encoding = request['format'].get('encoding', 'PROTO')
        self.sample_interval = request['format'].get('sample_interval', 10)
        self.stream_max = request['format'].get('stream_max', 0)
        self.time_delta = 0
        self.result = None
        self.event_triggered = False
        self._request = request

    def process_opfields(self, response):
        """Decode response and verify result.

        Decoder callback returns desired format of response.
        Verify callback returns verification of expected results.

        Args:
          response (proto.gnmi_pb2.Notification): Contains updates that
              have changes since last timestamp.
        """
        subscribe_resp = json_format.MessageToDict(response)
        updates = subscribe_resp['update']
        for update in updates['update']:
            resp = self.decode_response(update, self.namespace)
            if self.event_triggered:
                if resp:
                    if not self.returns:
                        self.log.error('No notification values to check')
                        self.result = False
                        self.stop()
                    else:
                        self.result = self.response_verify(resp, self.returns.copy())
                else:
                    self.log.error('No values in subscribe response')

    def run(self):
        """Check for inbound notifications."""
        t1 = datetime.now()
        self.log.info('\nSubscribe notification active\n{0}'.format(
            29 * '='
        ))
        try:
            for response in self.responses:
                self.log.info(response)
                if self.stopped():
                    self.time_delta = self.stream_max
                    self.log.info("Terminating notification thread")
                    break
                if self.stream_max:
                    t2 = datetime.now()
                    td = t2 - t1
                    self.time_delta = td.seconds
                    if td.seconds > self.stream_max:
                        self.stop()
                        break
                if response.HasField('sync_response'):
                    self.log.info('Subscribe syncing response')
                if response.HasField('update'):
                    self.log.info(
                        '\nSubscribe response:\n{0}\n{1}'.format(
                            19 * '=',
                            str(response)
                        )
                    )
                    self.process_opfields(response)

        except Exception as exc:
            msg = ''
            if hasattr(exc, 'details'):
                msg += 'details: ' + exc.details()
            if hasattr(exc, 'debug_error_string'):
                msg += exc.debug_error_string()
            if not msg:
                msg = str(exc)
            self.result = msg

    def stop(self):
        self.log.info("Stopping notification stream")
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()


class Gnmi(BaseConnection):
    """Session handling for gNMI connections.

    Can be used with pyATS same as yang.connector.Netconf is used or
    can be used as a standlone module.

    Methods:

    capabilities(): gNMI Capabilities.
    set(dict): gNMI Set.  Input is namespace, xpath/value pairs.
    get(dict): gNMI Get mode='STATE'. Input xpath/value pairs (value optional).
    get_config(dict): gNMI Get mode='CONFIG'. Input xpath/value pairs.
    subscribe(dict): gNMI Subscribe.  Input xpath/value pairs and format
    notify_wait(dict, callback): Notify subscibe thread that event occured, \
        "callback" must be a class with passed, and failed methods and a \
        result class containing "code" property.

    pyATS Examples:

    >>> from pyats.topology import loader
    >>> from yang.connector.gnmi import Gnmi
    >>> testbed=loader.load('testbed_native_test.yaml')
    >>> device=testbed.devices['ddmi-9500-2']
    >>> device.connect(alias='gnmi', via='yang2')
    >>> #####################
    >>> # Set/Get example   #
    >>> #####################
    >>> content={
    ... 'namespace': {'ios: 'http://cisco.com/ns/yang/Cisco-IOS-XE-native',
    ... 'ios-cdp': 'http://cisco.com/ns/yang/Cisco-IOS-XE-cdp'},
    ... 'nodes': [{'xpath': '/ios:native/ios:cdp/ios-cdp:holdtime',
    ... 'value': '10'}]
    ... }
    >>> device.gnmi.set(content)
    []
    >>> content['nodes'][0].pop('value')
    >>> device.gnmi.get(content)
    [{'update': [(10, '/native/cdp/holdtime')]}]
    >>> #####################
    >>> # Capabilities      #
    >>> #####################
    >>> resp=device.gnmi.capabilities()
    >>> resp.keys()
    dict_keys(['supportedModels', 'supportedEncodings', 'gNMIVersion'])

    Standalone Examples (pyATS not installed):

    >>> #####################
    >>> # Capabilities      #
    >>> #####################
    >>> from yang.connector.gnmi import Gnmi
    >>> kwargs={
    ... 'host':'172.23.167.122',
    ... 'port':'50051',
    ... 'root_certificate':'root.pem',
    ... 'username':'admin',
    ... 'password':'C!sco123',
    ... 'ssl_name_override':'ems.cisco.com'
    ... }
    >>> gnmi=Gnmi('iosxe', **kwargs)
    >>> resp = gnmi.capabilities()
    >>> #####################
    >>> # Set example       #
    >>> #####################
    >>> content={
    ... 'namespace': {'ios: 'http://cisco.com/ns/yang/Cisco-IOS-XE-native',
    ... 'ios-cdp': 'http://cisco.com/ns/yang/Cisco-IOS-XE-cdp'},
    ... 'nodes': [{'xpath': '/ios:native/ios:cdp/ios-cdp:holdtime',
    ... 'value': '10'}]
    ... }
    >>> resp = gnmi.set(content)
    >>> #####################
    >>> # Get mode='CONFIG' #
    >>> #####################
    >>> content={
    ... 'namespace': {'ios: 'http://cisco.com/ns/yang/Cisco-IOS-XE-native',
    ... 'ios-cdp': 'http://cisco.com/ns/yang/Cisco-IOS-XE-cdp'},
    ... 'nodes': [{'xpath': '/ios:native/ios:cdp/ios-cdp:holdtime'}]
    ... }
    >>> resp = gnmi.get_config({'content': content, 'returns': returns})
    >>> #####################
    >>> # Get mode='STATE' #
    >>> #####################
    >>> content={
    ... 'namespace': {'ios: 'http://cisco.com/ns/yang/Cisco-IOS-XE-native',
    ... 'ios-cdp': 'http://cisco.com/ns/yang/Cisco-IOS-XE-cdp'},
    ... 'nodes': [{'xpath': '/ios:native/ios:cdp/ios-cdp:holdtime'}]
    ... }
    >>> resp = gnmi.get({'content': content, 'returns': returns})
    >>> #####################
    >>> # Subscribe example #
    >>> #####################
    >>> content={
    ... 'format': {
    ...     'encoding': 'JSON',
    ...     'request_mode': 'STREAM',
    ...     'sample_interval': 5,
    ...     'stream_max': 15,
    ...     'sub_mode': 'SAMPLE'},
    ... }
    >>> # Add namespace, node xpath list similar to Get to the content
    >>> # Add list of expected return values to content
    >>> # Example of an expected return value:
    >>> content['returns']=[{
    ... 'datatype': 'string',
    ... 'id': 0,
    ... 'name': 'name',
    ... 'op': '==',
    ... 'selected': True,
    ... 'value': 'v4acl',
    ... 'xpath': '/acl/acl-sets/acl-set/name'}]
    >>> # Trigger an event that would kick off subscribe
    >>> # Call notify_wait() passing in a callback class (see notify_wait)

   """

    os_class_map = {
        None: None,
        "iosxr": "IOS XR",
        "nxos": "NX-OS",
        "iosxe": "IOS XE",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.device = kwargs.get('device')
        self.channel = None
        dev_args = self.connection_info
        if dev_args.get('protocol', '') != 'gnmi':
            msg = 'Invalid protocol {0}'.format(
                dev_args.get('protocol', '')
            )
            raise TypeError(msg)

        # Initialize ClientBuilder
        self.client_os = self.os_class_map.get(self.device.os, None)
        if self.device.os == 'iosxe':
            self.support_prefix = True
            self.json_ietf = True
        else:
            self.support_prefix = False
            self.json_ietf = False

        # ClientBuilder target is IP:Port
        target = dev_args.get('host') + ':' + str(dev_args.get('port'))
        builder = ClientBuilder(target).set_os(self.client_os)

        # Gather certificate settings
        root = dev_args.get('root_certificate')
        if not root:
            root = None
        if root and os.path.isfile(root):
            root = open(root, 'rb').read()
        chain = dev_args.get('certificate_chain')
        if not chain:
            chain = None
        if chain and os.path.isfile(chain):
            chain = open(chain, 'rb').read()
        private_key = dev_args.get('private_key', '')
        if not private_key:
            private_key = None
        if private_key and os.path.isfile(private_key):
            private_key = open(private_key, 'rb').read()
        if any((root, chain, private_key)):
            builder.set_secure(root, private_key, chain)
            builder.set_ssl_target_override(
                dev_args.get('ssl_name_override', '')
            )
            log.info("Connecting secure channel")
        else:
            builder._set_insecure()
            log.info("Connecting insecure channel")

        # Get/set credentials
        username = dev_args.get('username', '')
        password = dev_args.get('password', '')
        if not username or not password:
            creds = dev_args.get('credentials', '')
            if not creds:
                raise KeyError('No credentials found for testbed')
            if 'gnmi' not in creds:
                log.info('Credentials used from {0}'.format(
                    next(iter(creds))
                ))
            gnmi_uname_pwd = creds.get('')
            if not gnmi_uname_pwd:
                raise KeyError('No credentials found for gNMI')
            username = gnmi_uname_pwd.get('username', '')
            password = gnmi_uname_pwd.get('password', '')
            if not username or not password:
                raise KeyError('No credentials found for gNMI testbed')
            password = to_plaintext(password)
        builder.set_call_authentication(username, password)
        # builder.construct() connects grpc channel and returns client instance
        self.builder = builder
        self.gnmi = None

    active_notifications = {}

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device):
        if device:
            self._device = device

    @property
    def connected(self):
        """Return True if session is connected."""
        return self.gnmi

    def path_elem_to_xpath(self, path_elem, namespace={}, opfields=[]):
        """Convert a Path structure to an Xpath."""
        elems = path_elem.get('elem', [])
        xpath = []
        for elem in elems:
            name = elem.get('name', '')
            if name:
                for mod in namespace.values():
                    name = name.replace(mod + ':', '')
                xpath.append(name)
            key = elem.get('key', '')
            if key:
                for name, value in key.items():
                    for mod in namespace.values():
                        value = str(value).replace(mod + ':', '')
                    opfields.append((
                        value,
                        '/' + '/'.join(xpath) + '/' + name,

                    ))
        return '/' + '/'.join(xpath)

    def get_opfields(self, val, xpath_str, opfields=[], namespace={}):
        if isinstance(val, dict):
            for name, dict_val in val.items():
                opfields = self.get_opfields(
                    dict_val,
                    xpath_str + '/' + name,
                    opfields,
                    namespace
                )
        elif isinstance(val, list):
            for item in val:
                self.get_opfields(item, xpath_str, opfields, namespace)
        else:
            xpath_list = xpath_str.split('/')
            name = xpath_list.pop()
            for mod in namespace.values():
                name = name.replace(mod + ':', '')
            xpath_str = '/'.join(xpath_list)
            opfields.append((val, xpath_str + '/' + name))

        return opfields

    def decode_opfields(self, update={}, namespace={}):
        opfields = []
        xpath_str = self.path_elem_to_xpath(
            update.get('path', {}),
            namespace=namespace,
            opfields=opfields
        )
        if not xpath_str:
            log.error('Xpath not determined from response')
            return []
        # TODO: the val depends on the encoding type
        val = update.get('val', {}).get('jsonIetfVal', '')
        if not val:
            val = update.get('val', {}).get('jsonVal', '')
        if not val:
            log.error('{0} has no values'.format(xpath_str))
            return []
        json_val = base64.b64decode(val).decode('utf-8')
        update_val = json.loads(json_val)
        if not isinstance(update_val, (dict, list)):
            # Just one value returned
            opfields.append((update_val, xpath_str))
            return opfields
        else:
            # Reset opfields to avoid duplicates
            opfields = []
        if isinstance(update_val, dict):
            opfields = self.get_opfields(
                update_val,
                xpath_str,
                opfields,
                namespace
            )
        elif isinstance(update_val, list):
            for val_dict in update_val:
                opfields = self.get_opfields(
                    val_dict,
                    xpath_str,
                    opfields,
                    namespace
                )
        return opfields

    def decode_update(self, update={}, namespace={}):
        """Convert JSON return to python dict for display or processing."""
        val = update.get('val', {}).get('jsonIetfVal', '')
        if not val:
            val = update.get('val', {}).get('jsonVal', '')
        if not val:
            log.error('Update has no value')
            return []
        json_val = base64.b64decode(val).decode('utf-8')
        return json.loads(json_val)

    def decode_notification(self, response, namespace={}):
        """Decode a response from the google.protobuf into a dict."""
        if isinstance(response, dict):
            resp_dict = response
        else:
            resp_dict = json_format.MessageToDict(response)
        notifies = resp_dict.get('notification', [])
        ret_vals = []
        for notify in notifies:
            ret_val = {}
            time_stamp = notify.get('timestamp', '')
            # TODO: convert time_stamp from str nanoseconds since epoch time
            # to datetime
            for update in notify['update']:
                val_dict = self.decode_update(update, namespace=namespace)
                opfields = self.decode_opfields(update, namespace=namespace)
                if 'decode' not in ret_val:
                    ret_val['decode'] = [val_dict]
                else:
                    ret_val['decode'].append(val_dict)
                if 'update' not in ret_val:
                    ret_val['update'] = [opfields]
                else:
                    ret_val['update'].append(opfields)
            deletes = notify.get('delete', [])
            deleted = []
            for delete in deletes:
                xpath_str = self.path_elem_to_xpath(
                    delete.get('path', {}),
                    namespace
                )
                if xpath_str:
                    deleted.append(xpath_str)
            if deleted:
                ret_val['delete'] = deleted
            ret_vals.append(ret_val)
        return ret_vals

    def decode_capabilities(self, caps={}):
        return_caps = {
            'version': 'unknown',
            'encodings': 'unknown',
            'models': []
        }
        if not caps:
            return {}
        caps_dict = json_format.MessageToDict(caps)
        for key, value in caps_dict.items():
            if key == 'gNMIVersion':
                return_caps['version'] = value
            elif key == 'supportedEncodings':
                return_caps['encodings'] = value
            elif key == 'supportedModels':
                return_caps['models'] = value

        return return_caps

    def connect(self):
        # builder.construct() connects grpc channel and returns client
        if self.connected:
            return
        self.gnmi, self.channel = self.builder.construct(return_channel=True)
        resp = self.capabilities()
        if resp:
            log.info(
                '\ngNMI version: {0} supported encodings: {1}\n\n'.format(
                    resp.get('gNMIVersion', 'unknown'),
                    resp.get('supportedEncodings', 'unknown')
                ))
            log.info(banner('gNMI CONNECTED'))
        else:
            log.info(banner('gNMI Capabilities not returned'))
            self.disconnect()
            raise gNMIException('Connection not successful')

    def set(self, cmd):
        """Send any Set data command.

        Args:
          cmd (dict): Mapping to namespace, xpath/value. \
              {'namespace': '<prefix>': '<namespace>'}, \
              {'nodes': [{ \
                  'edit-op': '<netconf edit-config operation', \
                  'xpath': '<prefixed Xpath to resource', \
                  'value': <value to set resource to> \
              }]} \
        Returns:
          (dict): gNMI SetResponse
        """
        if not self.connected:
            self.connect()
        try:
            # Convert xpath to path element
            responses = []
            ns, configs, origin = xpath_util.xml_path_to_path_elem(cmd)
            prefix = xpath_util.get_prefix(origin) if self.support_prefix else None
            updates = configs.get('update')
            if len(updates) > 1:
                xpaths = []
                for update in updates:
                    xpath = next(iter(update.keys()))
                    xpaths.append(xpath)
                if os.path.commonprefix(xpaths):
                    updates = xpath_util.get_payload(updates)
            replaces = configs.get('replace')
            if len(replaces) > 1:
                xpaths = []
                for replace in replaces:
                    xpath = next(iter(replace.keys()))
                    xpaths.append(xpath)
                if os.path.commonprefix(xpaths):
                    replaces = xpath_util.get_payload(replaces)

            deletes = configs.get('delete')
            if updates or replaces:
                response = self.gnmi.set_json(
                    updates,
                    replaces,
                    ietf=self.json_ietf,
                    prefix=prefix
                )
                responses.append(response)
            if deletes:
                response = self.gnmi.delete_xpaths(deletes)
                responses.append(response)
            # Do fixup on response
            return responses
        except Exception as exc:
            msg = ''
            msg = exc.details() if hasattr(exc, 'details') else str(exc)
            log.error(banner('ERROR: {0}'.format(msg)))

    def configure(self, cmd):
        return self.set(cmd)

    def get(self, cmd, datatype='STATE'):
        """Send any Get data commmand.

        Args:
          cmd (dict): Mapping to namespace, xpath. \
              {{'namespace': '<prefix>': '<namespace>'}, \
               {'nodes': [{'xpath': '<prefixed Xpath to resource'}]} \
          datatype (str): [ ALL | STATE ] (default: STATE) \
        Returns:
          list: List of dict containing updates, replaces, deletes. \
                Updates and replaces are lists of value/xpath tuples. \
                    [{'updates': [(<value>, <xpath"), ...]}] \
                Deletes are a list of xpaths. \
                    [<xpath>,...] \
        """
        if not self.connected:
            self.connect()
        try:
            # Convert xpath to path element
            ns, msg, origin = xpath_util.xml_path_to_path_elem(
                cmd,
                self.support_prefix
            )
            resp = self.gnmi.get_xpaths(
                msg.get('get', []),
                data_type=datatype
            )
            log.info('\nGNMI response:\n{0}\n{1}'.format(15 * '=', str(resp)))
            # Do fixup on response
            response = self.decode_notification(resp, ns)
            # TODO: Do we need to send back deletes?
            return response
        except Exception as exc:
            msg = ''
            if hasattr(exc, 'details'):
                msg = exc.details()
            else:
                msg = str(exc)
            log.error(banner('ERROR: {0}'.format(msg)))
            return []

    def get_config(self, cmd):
        """Helper function for get.

        Args:
          cmd (dict): See get function.
        Returns:
          list: See get function.
        """
        return self.get(cmd, datatype='CONFIG')

    def execute(self, cmd):
        return self.get(cmd)

    def capabilities(self):
        """Retrieve capabilities from device.

        Returns:
          dict: All capability information.
        """
        if not self.connected:
            self.connect()
        try:
            response = json_format.MessageToDict(self.gnmi.capabilities())
            log.debug('\nDevice capabilities\n{0}{1}'.format(
                19 * '=', str(response))
            )
            return response
        except Exception as exc:
            msg = ''
            msg = exc.details() if hasattr(exc, 'details') else str(exc)
            log.error(banner('ERROR: {0}'.format(msg)))
            return {}

    def subscribe(self, cmd):
        """Subscribe to notification stream.

        Send a subscribe command followed by a "notify_wait()" call.

        The subscribe command sends a request and listens for responses.
        Another command is then executed that would trigger an event.
        Notify wait is then called to inform the notification thread that
        an event was trigger and begin evaluating the returned data against
        expected results.

        Args:
          cmd (dict): Contains:
            'format': { \
              'encoding': [JSON | JSON_IETF], \
              'request_mode': [STREAM | ONCE | POLL], \
              'sample_interval': seconds between sampling, \
              'stream_max': Maximun time to keep stream open in seconds, \
              'sub_mode': [ON_CHANGE | SAMPLE]}, \
            }
            'namespace': same as in get function
            'nodes': same as in get function
            'returns': list of expected data returned in notifications.
                       Example of expected data dict:
                        {'datatype': 'string',
                            'name': 'name',
                            'op': '==',
                            'selected': True,
                            'value': 'v4acl',
                            'xpath': '/acl/acl-sets/acl-set/name'}

        """
        if not self.connected:
            self.connect()
        try:
            # Convert xpath to path element
            ns, msg, origin = xpath_util.xml_path_to_path_elem(cmd)
            prefix = xpath_util.get_prefix(origin) if self.support_prefix else None
            format = cmd['format']
            subscribe_xpaths = msg['get']
            subscribe_response = self.gnmi.subscribe_xpaths(
                subscribe_xpaths,
                format.get('request_mode', 'STREAM'),
                format.get('sub_mode', 'SAMPLE'),
                format.get('encoding', 'JSON_IETF'),
                self.gnmi._NS_IN_S * format.get('sample_interval', 10),
                prefix=prefix
            )
            if format.get('request_mode', 'STREAM') == 'ONCE':
                # Do fixup in response
                response_verify = cmd.get('verifier')
                returns = cmd.get('returns')
                for response in subscribe_response:
                    if response.HasField('update'):
                        resp = json_format.MessageToDict(response)
                        update = resp.get('update')
                        opfields = self.decode_opfields(
                            update.get('update'),
                            ns
                        )
                        return response_verify(opfields, returns)
                    return False
            else:
                cmd['namespace'] = ns
                cmd['decode'] = self.decode_opfields
                subscribe_thread = GnmiNotification(
                    self,
                    subscribe_response,
                    **cmd
                )
                subscribe_thread.start()
                self.active_notifications[self] = subscribe_thread
                if cmd['format']['request_mode'] == 'ON_CHANGE':
                    subscribe_thread.event_triggered = True
                return True
        except Exception as exc:
            msg = ''
            msg = exc.details() if hasattr(exc, 'details') else str(exc)
            log.error(banner('ERROR: {0}'.format(msg)))

    def notify_wait(self, steps):
        """Activate notification thread and check results."""
        notifier = self.active_notifications.get(self)
        if notifier:
            if steps.result.code != 1:
                notifier.stop()
                del self.active_notifications[self]
                return
            notifier.event_triggered = True
            log.info(banner('NOTIFICATION EVENT TRIGGERED'))
            wait_for_sample = notifier.sample_interval - 1
            cntr = 1.0
            while cntr < float(notifier.stream_max):
                log.info('Listening for notifications from subscribe stream, {} seconds elapsed'.format(
                    cntr)
                )
                cntr += 1
                if notifier.result is not None and wait_for_sample <= 0:
                    notifier.stop()
                    if notifier.result is True:
                        steps.passed(
                            '\n' + banner('NOTIFICATION RESPONSE PASSED')
                        )
                    else:
                        steps.failed(
                            '\n' + banner('NOTIFICATION RESPONSE FAILED')
                        )
                    break
                sleep(1)
                wait_for_sample -= 1
            else:
                notifier.stop()
                steps.failed('\n' + banner('STREAM TIMED OUT WITHOUT RESPONSE'))

            if self in self.active_notifications:
                del self.active_notifications[self]

    def get_updates(self, response):
        """Notification check."""
        return self.decode_response(response)

    def disconnect(self):
        """Disconnect from SSH device."""
        if self.connected:
            if self.channel:
                self.channel.close()
            self.gnmi = None
            del self.channel

    def __enter__(self):
        """Establish a session using a Context Manager."""
        if not self.connected:
            self.connect()
        return self

    def __exit__(self, *args):
        """Gracefully close connection on Context Manager exit."""
        self.disconnect()



#!/bin/env python
""" Unit tests for the yang.connector cisco-shared package. """

import unittest
from ncclient import manager
from ncclient import transport
from ncclient.devices.default import DefaultDeviceHandler
from pyats.topology import loader
from pyats.connections import BaseConnection
from unittest.mock import Mock, patch
import yang.connector


class MyTransportSession():

    def __init__(self):
        pass

    def close(self):
        pass


class MySSHSession():

    def __init__(self):
        self._is_stopped = False
        self._connected = False
        self.transport = None

    @property
    def connected(self):
        return self._connected

    def connect(self, **kwargs):
        self._connected = True
        self.transport = MyTransportSession()

    def close(self):
        self._connected = False

    def send(self, message):
        pass

    def get_listener_instance(self, cls):
        pass

    def add_listener(self, listener):
        pass

    def is_alive(self):
        return True

class MySSHSession2():

    def __init__(self):
        self._is_stopped = False
        self._connected = False
        self.transport = None

    @property
    def connected(self):
        return self._connected

    def connect(self, **kwargs):
        if kwargs['username'] == 'admin' and kwargs['password'] == 'admin':
            self._connected = True
            self.transport = MyTransportSession()

    def close(self):
        self._connected = False

    def send(self, message):
        pass

    def get_listener_instance(self, cls):
        pass

    def add_listener(self, listener):
        pass

    def is_alive(self):
        return True

class MyRawRPC():

    def __init__(self, session=None, device_handler=None,
                       timeout=None, raise_mode=None):
        self._event = MyEvent()
        self._listener = MyRPCReplyListener()

    def _request(self, msg):
        return MyRPCReply()

class MyRPCReplyListener():

    def __init__(self):
        pass

    def register(self, id, rpc):
        pass

class MyRPCReply():

    def __init__(self):
        self.xml = '''
            <rpc-reply message-id="101"
             xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" >
            <data>
            <native xmlns="http://cisco.com/ns/yang/ned/ios">
            <version>16.3</version>
            </native>
            </data>
            </rpc-reply>
            '''

class MyCloseSession():

    def __init__(self, session, **kwargs):
        pass

    def request(self):
        return MyRPCReply()

class MyEvent():

    def __init__(self):
        pass

    def isSet(self):
        return True

    def wait(self, timeout):
        pass

class TestYang(unittest.TestCase):

    def setUp(self):
        self.yaml = \
            'devices:\n' \
            '    dummy:\n' \
            '        type: dummy_device\n' \
            '        connections:\n' \
            '            netconf:\n' \
            '                class: yang.connector.Netconf\n' \
            '                ip : "1.2.3.4"\n' \
            '                port: 830\n' \
            '                username: admin\n' \
            '                password: admin\n'

        self.testbed = loader.load(self.yaml)
        self.device = self.testbed.devices['dummy']
        self.nc_device = yang.connector.Netconf(device=self.device,
                                                alias='nc', via='netconf')

    def test_disconnect(self):
        self.nc_device._session = MySSHSession()
        self.nc_device.disconnect()
        generated_value = self.nc_device.connected
        expected_value = False
        self.assertEqual(generated_value, expected_value)

    def test_connect_1(self):
        self.nc_device._session = MySSHSession()
        self.nc_device.connect()
        generated_value = self.nc_device.connected
        expected_value = True
        self.assertEqual(generated_value, expected_value)

    def test_connect_2(self):
        self.nc_device._session = MySSHSession()
        self.nc_device.connect()
        generated_value = self.nc_device.connected
        expected_value = True
        self.assertEqual(generated_value, expected_value)

    def test_connect_3(self):
        nc_device = yang.connector.Netconf(device=self.device,
                                           alias='nc2',
                                           via='netconf',
                                           username='admin',
                                           password='admin')
        nc_device._session = MySSHSession2()
        nc_device.connect()
        generated_value = nc_device.connected
        expected_value = True
        self.assertEqual(generated_value, expected_value)

    def test_connect_4(self):
        nc_device = yang.connector.Netconf(device=self.device,
                                           alias='nc2',
                                           via='netconf',
                                           username='admi',
                                           password='admin')
        nc_device._session = MySSHSession2()
        nc_device.connect()
        generated_value = nc_device.connected
        expected_value = False
        self.assertEqual(generated_value, expected_value)

    @patch('yang.connector.netconf.RawRPC', new=MyRawRPC)
    def test_request(self):
        self.nc_device._session = MySSHSession()
        self.nc_device.connect()
        r = '''
             <rpc message-id="101"
              xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
             <get>
             <filter>
             <native xmlns="http://cisco.com/ns/yang/ned/ios">
             <version>
             </version>
             </native>
             </filter>
             </get>
             </rpc>
            '''
        generated_value = self.nc_device.request(r)
        expected_value = '''
            <rpc-reply message-id="101"
             xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" >
            <data>
            <native xmlns="http://cisco.com/ns/yang/ned/ios">
            <version>16.3</version>
            </native>
            </data>
            </rpc-reply>
            '''
        self.assertEqual(generated_value, expected_value)

    def test_rawrpc(self):
        from ncclient.operations.retrieve import GetReply

        h = DefaultDeviceHandler()
        self.rawrpc = yang.connector.netconf.RawRPC(session = transport.SSHSession(h),
                                            device_handler = h)
        self.rawrpc._event = MyEvent()
        self.rawrpc._session = MySSHSession()
        reply_raw = '''
            <rpc-reply message-id="101"
             xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" >
            <data>
            <native xmlns="http://cisco.com/ns/yang/ned/ios">
            <version>16.3</version>
            </native>
            </data>
            </rpc-reply>
            '''
        self.rawrpc._reply = GetReply(reply_raw)
        r = '''
             <rpc message-id="101"
              xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
             <get>
             <filter>
             <native xmlns="http://cisco.com/ns/yang/ned/ios">
             <version>
             </version>
             </native>
             </filter>
             </get>
             </rpc>
            '''
        generated_value = self.rawrpc._request(r).xml
        expected_value = '''
            <rpc-reply message-id="101"
             xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" >
            <data>
            <native xmlns="http://cisco.com/ns/yang/ned/ios">
            <version>16.3</version>
            </native>
            </data>
            </rpc-reply>
            '''
        self.assertEqual(generated_value, expected_value)

    def test_config(self):
        self.assertRaises(Exception,
                          self.nc_device.configure,
                          'logging console')

    def test_execute_1(self):
        manager.OPERATIONS = {
            "close_session": MyCloseSession,
        }
        self.nc_device.execute('close_session')
        generated_value = self.nc_device.connected
        expected_value = False
        self.assertEqual(generated_value, expected_value)

    def test_execute_2(self):
        self.assertRaises(Exception,
                          self.nc_device.execute,
                          'close')


if __name__ == '__main__':
    unittest.main()



import os
import re
import json
import logging
from six import string_types
from cisco_gnmi.client import proto

log = logging.getLogger(__name__)


def get_prefix(origin):
    # TODO: calculate a prefix instead of combining config?
    if origin == 'openconfig':
        # No prefix support for openconfig
        return None
    prefix_path = proto.gnmi_pb2.Path()
    return prefix_path


def combine_configs(payload, last_xpath, cfg):
    """Walking from end to finish, 2 xpaths merge, so combine them.
                               |--config
        |---last xpath config--|
    ----|                      |--config
        |
        |   pick these up -->  |--config
        |---this xpath config--|
                               |--config
    Parameters
    ----------
    payload: dict of partial payload
    last_xpath: last xpath that was processed
    xpath: colliding xpath
    config: dict of values associated to colliding xpath
    """
    xpath, config, is_key = cfg
    lp = last_xpath.split("/")
    xp = xpath.split("/")
    base = []
    top = ""
    for i, seg in enumerate(zip(lp, xp)):
        if seg[0] != seg[1]:
            top = seg[1]
            break
    base = "/" + "/".join(xp[i:])
    cfg = (base, config, False)
    extended_payload = {top: xpath_to_json([cfg])}
    payload.update(extended_payload)
    return payload


def xpath_to_json(configs, last_xpath="", payload={}):
    """Try to combine Xpaths/values into a common payload (recursive).

    Parameters
    ----------
    configs: tuple of xpath/value dict
    last_xpath: str of last xpath that was recusivly processed.
    payload: dict being recursively built for JSON transformation.

    Returns
    -------
    dict of combined xpath/value dict.
    """
    for i, cfg in enumerate(configs, 1):
        xpath, config, is_key = cfg
        if last_xpath and xpath not in last_xpath:
            # Branched config here     |---config
            #   |---last xpath config--|
            # --|                      |---config
            #   |---this xpath config
            payload = combine_configs(payload, last_xpath, cfg)
            return xpath_to_json(configs[i:], xpath, payload)
        xpath_segs = xpath.split("/")
        xpath_segs.reverse()
        for seg in xpath_segs:
            if not seg:
                continue
            if payload:
                if is_key:
                    if seg in payload:
                        if isinstance(payload[seg], list):
                            payload[seg].append(config)
                        elif isinstance(payload[seg], dict):
                            payload[seg].update(config)
                    else:
                        payload.update(config)
                        payload = {seg: [payload]}
                else:
                    config.update(payload)
                    payload = {seg: config}
                return xpath_to_json(configs[i:], xpath, payload)
            else:
                if is_key:
                    payload = {seg: [config]}
                else:
                    payload = {seg: config}
                return xpath_to_json(configs[i:], xpath, payload)
    return payload


# Pattern to detect keys in an xpath
RE_FIND_KEYS = re.compile(r"\[.*?\]")


def get_payload(configs):
    """Common Xpaths were detected so try to consolidate them.

    Parameter
    ---------
    configs: list of {xpath: {name: value}} dicts
    """
    # Number of updates are limited so try to consolidate into lists.
    xpaths_cfg = []
    first_key = set()
    # Find first common keys for all xpaths_cfg of collection.
    for config in configs:
        xpath = next(iter(config.keys()))

        # Change configs to tuples (xpath, config) for easier management
        xpaths_cfg.append((xpath, config[xpath]))

        xpath_split = xpath.split("/")
        for seg in xpath_split:
            if "[" in seg:
                first_key.add(seg)
                break

    # Common first key/configs represents one GNMI update
    updates = []
    for key in first_key:
        update = []
        remove_cfg = []
        for config in xpaths_cfg:
            xpath, cfg = config
            if key in xpath:
                update.append(config)
            else:
                for k, v in cfg.items():
                    if '[{0}="{1}"]'.format(k, v) not in key:
                        break
                else:
                    # This cfg sets the first key so we don't need it
                    remove_cfg.append((xpath, cfg))
        if update:
            for upd in update:
                # Remove this config out of main list
                xpaths_cfg.remove(upd)
            for rem_cfg in remove_cfg:
                # Sets a key in update path so remove it
                xpaths_cfg.remove(rem_cfg)
            updates.append(update)
            break

    # Add remaining configs to updates
    if xpaths_cfg:
        updates.append(xpaths_cfg)

    # Combine all xpath configs of each update if possible
    xpaths = []
    compressed_updates = []
    for update in updates:
        xpath_consolidated = {}
        config_compressed = []
        for seg in update:
            xpath, config = seg
            if xpath in xpath_consolidated:
                xpath_consolidated[xpath].update(config)
            else:
                xpath_consolidated[xpath] = config
                config_compressed.append((xpath, xpath_consolidated[xpath]))
                xpaths.append(xpath)

        # Now get the update path for this batch of configs
        common_xpath = os.path.commonprefix(xpaths)
        cfg_compressed = []
        keys = []

        # Need to reverse the configs to build the dict correctly
        config_compressed.reverse()
        compressed_count = 0
        for seg in config_compressed:
            is_key = False
            prepend_path = ""
            xpath, config = seg
            end_path = xpath[len(common_xpath):]
            if not end_path:
                prepend_path = common_xpath
            elif end_path.startswith("["):
                # Don't start payload with a list
                tmp = common_xpath.split("/")
                prepend_path = "/" + tmp.pop()
                common_xpath = "/".join(tmp)
            end_path = prepend_path + end_path

            # Building json, need to identify configs that set keys
            for key in keys:
                if [k for k in config.keys() if k in key]:
                    is_key = True
            keys += re.findall(RE_FIND_KEYS, end_path)
            cfg_compressed.append((end_path, config, is_key))
            compressed_count += 1

        update = (common_xpath, cfg_compressed)
        compressed_updates.append(update)

    updates = []
    if compressed_count == 1:
        common_xpath, cfg = compressed_updates[0]
        xpath, payload, is_key = cfg[0]
        updates.append({xpath: payload})
    else:
        for update in compressed_updates:
            common_xpath, cfgs = update
            payload = xpath_to_json(cfgs)
            updates.append({common_xpath: payload})
    return updates


def xml_path_to_path_elem(request, prefix=False):
    """Convert XML Path Language 1.0 Xpath to gNMI Path/PathElement.

    Modeled after YANG/NETCONF Xpaths.

    References:
    * https://www.w3.org/TR/1999/REC-xpath-19991116/#location-paths
    * https://www.w3.org/TR/1999/REC-xpath-19991116/#path-abbrev
    * https://tools.ietf.org/html/rfc6020#section-6.4
    * https://tools.ietf.org/html/rfc6020#section-9.13
    * https://tools.ietf.org/html/rfc6241

    Parameters
    ---------
    request: dict containing request namespace and nodes to be worked on.
        namespace: dict of <prefix>: <namespace>
        nodes: list of dict
                <xpath>: Xpath pointing to resource
                <value>: value to set resource to
                <edit-op>: equivelant NETCONF edit-config operation

    Returns
    -------
    tuple: namespace_modules, message dict, origin
        namespace_modules: dict of <prefix>: <module name>
            Needed for future support.
        message dict: 4 lists containing possible updates, replaces,
            deletes, or gets derived form input nodes.
        origin str: DME, device, or openconfig
    """

    paths = []
    message = {
        "update": [],
        "replace": [],
        "delete": [],
        "get": [],
    }
    if "nodes" not in request:
        # TODO: raw rpc?
        return paths
    else:
        namespace_modules = {}
        origin = "DME"
        for prefix, nspace in request.get("namespace", {}).items():
            if "/Cisco-IOS-" in nspace:
                module = nspace[nspace.rfind("/") + 1 :]
            elif "/cisco-nx" in nspace:  # NXOS lowercases namespace
                module = "Cisco-NX-OS-device"
            elif "/openconfig.net" in nspace:
                module = "openconfig-"
                module += nspace[nspace.rfind("/") + 1 :]
            elif "urn:ietf:params:xml:ns:yang:" in nspace:
                module = nspace.replace("urn:ietf:params:xml:ns:yang:", "")
            if module:
                namespace_modules[prefix] = module

        for node in request.get("nodes", []):
            if "xpath" not in node:
                log.error("Xpath is not in message")
            else:
                xpath = node["xpath"]
                value = node.get("value", "")
                edit_op = node.get("edit-op", "")

                for pfx, mod in namespace_modules.items():
                    if pfx not in xpath:
                        continue
                    if "Cisco-IOS-" in mod:
                        origin = 'rfc7951'
                        mod += ":"
                    elif 'openconfig' in mod:
                        origin = 'openconfig'
                        mod = ''
                    elif 'Cisco-NX-OS' in mod:
                        origin = 'device'
                        mod = ''
                    # Adjust prefixes of xpaths
                    xpath = xpath.replace(pfx + ":", mod)
                    if isinstance(value, string_types):
                        value = value.replace(pfx + ":", mod)

                if edit_op:
                    if edit_op in ["create", "merge", "replace"]:
                        xpath_lst = xpath.split("/")
                        name = xpath_lst.pop()
                        xpath = "/".join(xpath_lst)
                        if edit_op == "replace":
                            if not message["replace"]:
                                message["replace"] = [{xpath: {name: value}}]
                            else:
                                message["replace"].append({xpath: {name: value}})
                        else:
                            if not message["update"]:
                                message["update"] = [{xpath: {name: value}}]
                            else:
                                message["update"].append({xpath: {name: value}})
                    elif edit_op in ["delete", "remove"]:
                        if message["delete"]:
                            message["delete"].add(xpath)
                        else:
                            message["delete"] = set(xpath)
                else:
                    message["get"].append(xpath)
    return namespace_modules, message, origin


if __name__ == "__main__":
    from pprint import pprint as pp

    request = {
        "namespace": {"oc-acl": "http://openconfig.net/yang/acl"},
        "nodes": [
            {
                "value": "testacl",
                "xpath": "/oc-acl:acl/oc-acl:acl-sets/oc-acl:acl-set/name",
                "edit-op": "merge",
            },
            {
                "value": "ACL_IPV4",
                "xpath": "/oc-acl:acl/oc-acl:acl-sets/oc-acl:acl-set/type",
                "edit-op": "merge",
            },
            {
                "value": "10",
                "xpath": '/oc-acl:acl/oc-acl:acl-sets/oc-acl:acl-set[name="testacl"][type="ACL_IPV4"]/oc-acl:acl-entries/oc-acl:acl-entry/oc-acl:sequence-id',
                "edit-op": "merge",
            },
            {
                "value": "20.20.20.1/32",
                "xpath": '/oc-acl:acl/oc-acl:acl-sets/oc-acl:acl-set[name="testacl"][type="ACL_IPV4"]/oc-acl:acl-entries/oc-acl:acl-entry[sequence-id="10"]/oc-acl:ipv4/oc-acl:config/oc-acl:destination-address',
                "edit-op": "merge",
            },
            {
                "value": "IP_TCP",
                "xpath": '/oc-acl:acl/oc-acl:acl-sets/oc-acl:acl-set[name="testacl"][type="ACL_IPV4"]/oc-acl:acl-entries/oc-acl:acl-entry[sequence-id="10"]/oc-acl:ipv4/oc-acl:config/oc-acl:protocol',
                "edit-op": "merge",
            },
            {
                "value": "10.10.10.10/32",
                "xpath": '/oc-acl:acl/oc-acl:acl-sets/oc-acl:acl-set[name="testacl"][type="ACL_IPV4"]/oc-acl:acl-entries/oc-acl:acl-entry[sequence-id="10"]/oc-acl:ipv4/oc-acl:config/oc-acl:source-address',
                "edit-op": "merge",
            },
            {
                "value": "DROP",
                "xpath": '/oc-acl:acl/oc-acl:acl-sets/oc-acl:acl-set[name="testacl"][type="ACL_IPV4"]/oc-acl:acl-entries/oc-acl:acl-entry[sequence-id="10"]/oc-acl:actions/oc-acl:config/oc-acl:forwarding-action',
                "edit-op": "merge",
            },
        ],
    }
    modules, message, origin = xml_path_to_path_elem(request)
    pp(modules)
    pp(message)
    pp(origin)
    """
    # Expected output
    =================
    {'oc-acl': 'openconfig-acl'}
    {'delete': [],
    'get': [],
    'replace': [],
    'update': [{'/acl/acl-sets/acl-set': {'name': 'testacl'}},
                {'/acl/acl-sets/acl-set': {'type': 'ACL_IPV4'}},
                {'/acl/acl-sets/acl-set[name="testacl"][type="ACL_IPV4"]/acl-entries/acl-entry': {'sequence-id': '10'}},
                {'/acl/acl-sets/acl-set[name="testacl"][type="ACL_IPV4"]/acl-entries/acl-entry[sequence-id="10"]/ipv4/config': {'destination-address': '20.20.20.1/32'}},
                {'/acl/acl-sets/acl-set[name="testacl"][type="ACL_IPV4"]/acl-entries/acl-entry[sequence-id="10"]/ipv4/config': {'protocol': 'IP_TCP'}},
                {'/acl/acl-sets/acl-set[name="testacl"][type="ACL_IPV4"]/acl-entries/acl-entry[sequence-id="10"]/ipv4/config': {'source-address': '10.10.10.10/32'}},
                {'/acl/acl-sets/acl-set[name="testacl"][type="ACL_IPV4"]/acl-entries/acl-entry[sequence-id="10"]/actions/config': {'forwarding-action': 'DROP'}}]}
    'openconfig'
    """



#! /usr/bin/env python

"""Setup file for yang package

See:
    https://packaging.python.org/en/latest/distributing.html
"""

import os
import re
import sys
import shlex
import unittest
import subprocess
from setuptools import setup, find_packages, Command
from setuptools.command.test import test

pkg_name = 'yang.connector'
pkg_path = '/'.join(pkg_name.split('.'))

class CleanCommand(Command):
    '''Custom clean command

    cleanup current directory:
        - removes build/
        - removes src/*.egg-info
        - removes *.pyc and __pycache__ recursively

    Example
    -------
        python setup.py clean

    '''

    user_options = []
    description = 'CISCO SHARED : Clean all build artifacts'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./src/*.egg-info')
        os.system('find . -type f -name "*.pyc" | xargs rm -vrf')
        os.system('find . -type d -name "__pycache__" | xargs rm -vrf')

class TestCommand(Command):
    user_options = []
    description = 'CISCO SHARED : Run unit tests against this package'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # where the tests are (relative to here)
        tests = os.path.join('src', pkg_path, 'tests')

        # call unittests
        sys.exit(unittest.main(
            module = None,
           argv = ['python -m unittest', 'discover', tests],
           failfast = True))

class BuildAndPreviewDocsCommand(Command):
    user_options = []
    description = 'CISCO SHARED : Build and privately distribute ' \
        'Sphinx documentation for this package'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        user = os.environ['USER']
        sphinx_build_cmd = "sphinx-build -b html -c ../../docs " \
            "-d ./__build__/documentation/doctrees docs/ ./__build__/documentation/html"
        target_dir = "/users/{user}/WWW/cisco_shared/{pkg_name}".\
            format(user = user, pkg_name = pkg_path)
        mkdir_cmd = "mkdir -p {target_dir}".format(target_dir=target_dir)
        rsync_cmd = "rsync -rvc ./__build__/documentation/ {target_dir}".\
            format(target_dir=target_dir)
        try:
            ret_code = subprocess.call(shlex.split(mkdir_cmd))
            if not ret_code:
                ret_code = subprocess.call(shlex.split(sphinx_build_cmd))
                if not ret_code:
                    ret_code = subprocess.call(shlex.split(rsync_cmd))
                    #print("\nYou may preview the documentation at the following URL:")
                    #print("http://wwwin-home.cisco.com/~{user}/cisco_shared/{pkg_name}/html".\
                    #    format(user=user, pkg_name=pkg_path))
                    sys.exit(0)
            sys.exit(1)
        except Exception:
            sys.exit(1)


def read(*paths):
    '''read and return txt content of file'''
    with open(os.path.join(os.path.dirname(__file__), *paths)) as fp:
        return fp.read()


def find_version(*paths):
    '''reads a file and returns the defined __version__ value'''
    version_match = re.search(r"^__version__ ?= ?['\"]([^'\"]*)['\"]",
                              read(*paths), re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# launch setup
setup(
    name = pkg_name,
    version = find_version('src', pkg_path, '__init__.py'),

    # descriptions
    description =  'YANG defined interface API protocol connector',
    long_description = 'Client capable of NETCONF and gNMI protocol',

    # the package's documentation page.
    url = 'https://github.com/CiscoTestAutomation/yang.git',

    # author details
    author = 'Jonathan Yang',
    author_email = 'yuekyang@cisco.com',
    maintainer_email =  'yang-python@cisco.com',

    # project licensing
    license = 'Apache 2.0',

    platforms =  ['Linux',],

    # see https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Testing',
    ],

    # uses namespace package
    namespace_packages = ['yang'],

    # project keywords
    keywords = 'pyats cisco-shared',

    # project packages
    packages = find_packages(where = 'src'),

    # project directory
    package_dir = {
        '': 'src',
    },

    # additional package data files that goes into the package itself
    package_data = {'':['README.rst']},

    # Standalone scripts
    scripts = [
    ],

    # console entry point
    entry_points = {
    },

    # package dependencies
    install_requires =  [
        'paramiko >= 1.15.1',
        'lxml >= 3.3.0',
        'ncclient >= 0.6.6, < 0.6.10',
        'grpcio <= 1.36.1',
        'cisco-gnmi >= 1.0.13, < 2.0.0',
    ],

    # any additional groups of dependencies.
    # install using: $ pip install -e .[dev]
    extras_require = {
        'dev': ['coverage',
                'restview',
                'Sphinx',
                'sphinxcontrib-napoleon',
                'sphinx-rtd-theme'],
    },

    # any data files placed outside this package.
    # See: http://docs.python.org/3.4/distutils/setupscript.html
    # format:
    #   [('target', ['list', 'of', 'files'])]
    # where target is sys.prefix/<target>
    data_files = [],

    # custom commands for setup.py
    cmdclass = {
        'clean': CleanCommand,
        'test': TestCommand,
        'docs': BuildAndPreviewDocsCommand,
    },

    # non zip-safe (never tested it)
    zip_safe = False,
)



__author__    = 'Jonathan Yang <yuekyang@cisco.com>'
__date__      = 'July 22 2016'
__contact__   = 'yang-python@cisco.com'
__copyright__ = 'Cisco System, Inc. Cisco Confedential'

__import__('pkg_resources').declare_namespace(__name__)
